from datetime import date, timedelta
from pandas import concat, DataFrame, date_range, read_csv

from Library.error_functions import MAE, MAPE


def write_to_file(y_test, y_hat, model_name, training_time,
                  update_time, rolling_forecast, validation):    
    if rolling_forecast:
        rolling = 'rolling'
    else:
        rolling = ''
    
    y_hat.to_csv('Saved models\\' + model_name + ' ' + validation
                       + ' ' + rolling + '.csv')
    
    file = open('Saved Models\\Transcript.txt', 'a') 
    file.write('\n' + model_name + ' ' + validation + ' ' + str(rolling)
               + '\t  MAPE: ' + str(round(MAPE(y_test.values, y_hat.values), 2))
               + ' / MAE: ' + str(round(MAE(y_test.values, y_hat.values), 2))
               + ' / Training time: ' + str(int(training_time)) + 's'
               + ' / Update time: ' + str(round(update_time, 3)) + 's')
    file.close()
    
    
def load_redictions(model, test_year, validation='Expanding'):
    if model == 'Direct SVR' or model == 'Direct Lin_reg':
        updates = 'retraining'
    else:
        updates = 'rolling'
    
    return read_csv('Saved Models/'+model+' '+str(test_year)+' '+validation+' '+updates+'.csv',
                    index_col=0, names=[0])[0]
    
    
def load_all_predictions_for_year(dw, test_year, validation='Expanding'):
    power = dw.get('power')['power'] 
    power = power[1:][(power.index.year == test_year)]
    predictions = dw.get('Kalman')
    predictions['NTE Kalman'] = predictions['Kalman']
    del predictions['Kalman']
    predictions = predictions[1:][(predictions.index.year == test_year)[:-1]]
    
    for model in ['Direct Kalman', 'Direct Lin_reg', 'Direct MLP',
                  'Direct SVR', 'Elman Encoder-decoder',
                  'LSTM Encoder-decoder', 'MIMO MLP', 'SMSO MLP']:
        if model == 'Direct SVR' or model == 'Direct Lin_reg':
            updates = 'retraining'
        else:
            updates = 'rolling'
        
        series = read_csv('Saved Models/'+model+' '+str(test_year)+' '
                          +validation+' '+updates+'.csv', index_col=0, names=[0])[0]
        series.index = power.index
        predictions[model] = series
    
    predictions['Ensemble'] = (1/3)*(predictions['MIMO MLP'] 
    + predictions['SMSO MLP'] + predictions['Direct Kalman'])

    return predictions


def series_to_supervised(data, n_in=1, n_out=1, dropnan=True):
	n_vars = 1 if type(data) is list else data.shape[1]
	df = DataFrame(data)
	cols, names = list(), list()
    
	for i in range(n_in, 0, -1):
		cols.append(df.shift(i))
		names += [('var%d(t-%d)' % (j+1, i)) for j in range(n_vars)]
        
	for i in range(0, n_out):
		cols.append(df.shift(-i))
		if i == 0:
			names += [('var%d(t)' % (j+1)) for j in range(n_vars)]
		else:
			names += [('var%d(t+%d)' % (j+1, i)) for j in range(n_vars)]
            
	a = concat(cols, axis=1)
	a.columns = names
	if dropnan:
		a.dropna(inplace=True)
	return a

    
def generate_holidays(start_date, end_date):
    EASTER = {2011 : [4,24], 2012 : [4,8], 2013 : [3,31],
                       2014 : [4,20], 2015 : [4,5], 2016 : [3,27],
                       2017 : [4,16]}
    EASTER_OFFSETS = [-3, -2, 0, 1, 39, 49, 50]
    
    holidays = []
    for year in range(2011, 2017+1):
        holidays.append(date(year, 1, 1))
        holidays.append(date(year, 5, 1))
        holidays.append(date(year, 5, 17))
        holidays.append(date(year, 12, 25))
        holidays.append(date(year, 12, 26))
        
        pentecoast = date(year, EASTER[year][0], EASTER[year][1])
        for offset in EASTER_OFFSETS:
            holidays.append(pentecoast + timedelta(days=offset))
            
    holidays.sort()
    
    first = 0
    last = len(holidays)
    while first < last:
        if holidays[first] >= start_date:
            break
        else:
            first += 1
    while last > 0:
        if holidays[last-1] <= end_date:
            break
        else:
            last -= 1
    
    holidays = holidays[first:last]
            
    return holidays


def to_sequential(y_hat):
    predictions = []
    
    for index, row in y_hat.iterrows():
        for i in range(0, 24):
            predictions.append(row[i])
            
    begin_time = y_hat.index[0]
    end_time = y_hat.index[-1]
    
    # Daylight savings time was invented by the devil
    if end_time.date() in [date(2011, 3, 27), date(2012, 3, 25), date(2013, 3, 31),
                       date(2014, 3, 30), date(2015, 3, 29),
                       date(2016, 3, 27), date(2017, 3, 26)]:
        end_time += timedelta(hours=y_hat.shape[1]-2)
    elif end_time.date() in [date(2011, 10, 30), date(2012, 10, 28), date(2013, 10, 27),
                   date(2014, 10, 26), date(2015, 10, 25), 
                   date(2016, 10, 30), date(2017, 10, 29)]:
        end_time += timedelta(hours=y_hat.shape[1])
    else:
        end_time += timedelta(hours=y_hat.shape[1]-1)
        
    indexes = date_range(begin_time, end_time, freq='1h')
    for i in range(1, len(indexes)):
        if ((indexes[i].hour - indexes[i-1].hour) == 2):
            predictions = predictions[:i] + predictions[i+1:]
        if (indexes[i].hour == indexes[i-1].hour):
            predictions = predictions[:i] + [predictions[i]] + predictions[i:]
            
            
    y_hat_sequential = DataFrame(predictions)
    y_hat_sequential.index = indexes
    
    return y_hat_sequential[0]